/**
 *
 *	(c) Copyright Schlumberger Industries , created 1998.
 *	All use, disclosure, and/or
 *	reproduction is prohibited unless authorized in writing.
 *	All Rights Reserved.
 *	Performed by Schlumberger.
 *
 * Handle the display text of the rolling banner
 *
 * @author F. THIRIET, adapted my R. Kehr
 * @version V1.0
 * Date: 10/01/98, 11/01/20000
 * File: DisplayMessage.java
 *
*/

import javacard.framework.*;
import javacardx.framework.*;
import sim.toolkit.*;

public class DisplayMessage extends Page {

  final short MAXMSGSIZE = 256;
  byte messageBuf[] = new byte[MAXMSGSIZE];
	short lastChar;

	ProactiveElement message;
  byte messageProps = (byte)(HIGH_PRIORITY) | (byte)CLR_BY_USER;

	public DisplayMessage() {
  }

  public void open() {
    messageBuf[0] = (byte) 0xF4;
    lastChar = (short) 0x01;
  }

  public void setByte( byte b, short index) {
    messageBuf[index] = b;
  }

  public void setLength( short length ) {
    lastChar = length;
  }

  public void writeMessage( byte[] message ) {
    Util.arrayCopyNonAtomic( message, (short)0, messageBuf, lastChar, (short)message.length );
		lastChar = (short) ((short)lastChar + (short)message.length);
	}

  public void writeMessage( byte[] message, short len ) {
    Util.arrayCopyNonAtomic( message, (short)0, messageBuf, lastChar, len );
		lastChar = (short) ((short)lastChar + (short)len);
	}

	public void writeMessage( byte[] message, short start, short len ) {
    Util.arrayCopyNonAtomic( message, start, messageBuf, lastChar, len );
		lastChar  = (short) ((short)lastChar + (short)len);
	}

  public void writeByte( byte b ) {
    messageBuf[lastChar] = b;
    lastChar = (short) (lastChar + 1);
  }

  public void writeByteHex( byte b ) {
    Support.writeByte( b, messageBuf, lastChar );
    lastChar = (short) (lastChar + 2);
  }

  public void writeShortHex( short s ) {
    Support.writeShort( s, messageBuf, lastChar );
    lastChar = (short) (lastChar + 4);
  }

  public void writeDigit(byte b){
    Support.writeInvByte(b , messageBuf, lastChar);
    if ((byte)((b >> 4) & 15) == (byte) 15)
      lastChar = (short) (lastChar +1);
    else
      lastChar = (short) (lastChar +2);
  }

	public void setProperties( byte props ) {
		messageProps = props;
	}

  public void show() {
    buildCommand();
    ProactiveCommand.executeCommand();
  }
  
	public byte buildCommand() {
    if (message == null) {
		  message = new ProactiveElement(TEXT_STRING_TAG, MAXMSGSIZE);
		  APDU.waitExtension();
    }
		// Copy the string to the Text pro-active element
		Util.arrayCopyNonAtomic(messageBuf, (short)0, message.tlv, (short)2, lastChar);
		APDU.waitExtension();
    message.setLength( lastChar );
		APDU.waitExtension();
		ProactiveCommand.displayText((byte)1, messageProps, message);
		return ST_SUCCESS;
	}

/**
* Get the Terminal response from the previous command
* @return the next page for execution (in this case null or Help page)
*/
	public Page processResponse(){
		return super.processResponse();
	}
}

