// mandatory packages for STK applets
import javacard.framework.*;
import javacardx.framework.*;
import sim.toolkit.*;

public class WebSIM extends Applet implements ST, Constant, GSMCharacter
{
  short tp_SM_SD;
  short smsLength;

  //ProactiveElement req;
  ProactiveElement err;
  ProactiveElement about;
  ProactiveElement[] menuItem;

  public DisplayMessage displayMessage;
  public SelectItem selectItem;
  public SendSMS smsMessage;
  public Tone playTone;
  public WebServer server;
  public Input input;
  public Menu menu;

  public Page	currentPage;

  final static short MAX_PE_ELTS = 6;
  public ProactiveElement[] peConfig;

  final static short MAX_STACK_SIZE = (short) 8;
  Action[] nextActionStack;
  byte[] nextActionIdStack;
  short nextActionSP = (short) 0;


  public WebSIM () {
    APDU.waitExtension();
    register();
  }

  public void process (APDU apdu){

		switch (APDU.getBuffer()[ISO.OFFSET_INS]) {

			case APDU_ENVELOPE:

				if (APDU.getBuffer()[ISO.OFFSET_CDATA] == MENU_SELECTION) {
          APDU.waitExtension();
          currentPage = null;
          menu.process( this);
				} else if (APDU.getBuffer()[ISO.OFFSET_CDATA] == SMS_PP_DOWNLOAD) {
          currentPage = null;
          computeSMSOffsets();
          server.process( this, APDU.getBuffer(), tp_SM_SD, (short) smsLength );
				}
				break;

			case APDU_UPDATE_RECORD:
        currentPage = null;
        computeSMSOffsets();
        server.process( this, APDU.getBuffer(), tp_SM_SD, (short) smsLength );
   			break;

			case APDU_TERMINALRESPONSE:

				if (currentPage != null) {
          currentPage = currentPage.processResponse();
				  APDU.waitExtension(); // Restart timout
        }
 				if (currentPage != null){ // If a following page
					if (currentPage.buildCommand() == ST.SUCCESS) { // Prepare command for Fetch
						APDU.waitExtension(); // Restart timout
						ProactiveCommand.executeCommand(); // Fetch next page
					}
				} else {
          // If there is something on the call stack...
          //if (nextActionSP > (short)0) {
          //  APDU.waitExtension();
          //  nextActionSP = (short) ((short)nextActionSP - (short)1);
          //  nextActionStack[nextActionSP].restart( this, nextActionIdStack[nextActionSP] );
          //
          if (nextActionStack[0] != null) {
            nextActionStack[0].restart( this, nextActionIdStack[0] );
          }
        }
				break;

			case APDU_CONFIGURE : // Configure the text part of the pro-active element
        if (peConfig == null) {
          peConfig = new ProactiveElement[ MAX_PE_ELTS ];
          APDU.waitExtension();
          initialize( apdu );
        }
				ProactiveServices.configure( apdu, peConfig );
				break;

			case APDU_STATUS:
				break;
    }
  }

  public void pushAction( Action action, byte identifier ) {
    //nextActionStack[nextActionSP]   = action;
    //nextActionIdStack[nextActionSP] = identifier;
    //nextActionSP = (short) ((short)nextActionSP + (short)1);
    nextActionStack[0] = action;
    nextActionIdStack[0] = identifier;
  }

	public boolean select() {
	  return true;
	}

  public void initialize( APDU apdu ) {

    displayMessage = new DisplayMessage();
    APDU.waitExtension();

    selectItem = new SelectItem();
    APDU.waitExtension();

    smsMessage = new SendSMS();
    APDU.waitExtension();

    playTone = new Tone( apdu );
    APDU.waitExtension();

    input = new Input();
    APDU.waitExtension();

    menu = new Menu();
    APDU.waitExtension();

    server = new WebServer();
    APDU.waitExtension();

    nextActionStack   = new Action[MAX_STACK_SIZE];
    nextActionIdStack = new byte[MAX_STACK_SIZE];
    APDU.waitExtension();

    // Configures the proactive text elements
    // you have to initialize the three text elements
    // with the Applet Manager.
    // Create three TEXT_STRING Elements (WebSIM_req, WebSIM_about, WebSIM_err) and
    // set the value;
    // the peConfig Index is equal to the id in the configure dialog
    /*req = new ProactiveElement(TEXT_STRING_TAG, (short)50);
    APDU.waitExtension();
    peConfig[(short)0] = req;
    */
    about = new ProactiveElement(TEXT_STRING_TAG, (short)50);
    APDU.waitExtension();
    peConfig[(short)0] = about;

    err = new ProactiveElement(TEXT_STRING_TAG, (short)50);
    APDU.waitExtension();
    peConfig[(short)1] = err;

    menuItem = new ProactiveElement[5];

    menuItem[1] = new ProactiveElement(ITEM_TAG, (short) 32);
    APDU.waitExtension();
    peConfig[(short)2] = menuItem[1];

    menuItem[2] = new ProactiveElement(ITEM_TAG, (short) 32);
    APDU.waitExtension();
    peConfig[(short)3] = menuItem[2];

    menuItem[3] = new ProactiveElement(ITEM_TAG, (short) 32);
    APDU.waitExtension();
    peConfig[(short)4] = menuItem[3];

    menuItem[4] = new ProactiveElement(ITEM_TAG, (short) 32);
    APDU.waitExtension();
    peConfig[(short)5] = menuItem[4];

    APDU.waitExtension();
  }

  private void computeSMSOffsets() {
    short tpduOff;
    short tp_UD;
    short tp_UDL;
    short tp_SM_CPL;
    short tp_SM_CHL;
    short tp_SM_CPL_val;
    short tp_SM_CHL_val;

    APDU.waitExtension();
    tpduOff   = ProactiveServices.getTPDUSmsDownload();
    tp_UD     = ProactiveServices.getTP_FieldInSms(APDU.getBuffer(), tpduOff, TP_UD);
    tp_UDL    = ProactiveServices.getTP_FieldInSms(APDU.getBuffer(), tpduOff, TP_UDL);
    tp_SM_SD  = ProactiveServices.getTP_FieldInSms(APDU.getBuffer(), tpduOff, SM_SD);
    tp_SM_CPL = ProactiveServices.getTP_FieldInSms(APDU.getBuffer(), tpduOff, SM_CPL);
    tp_SM_CHL = ProactiveServices.getTP_FieldInSms(APDU.getBuffer(), tpduOff, SM_CHL);
    APDU.waitExtension();

    tp_SM_CPL_val = Util.getShort( APDU.getBuffer(), tp_SM_CPL );
    tp_SM_CHL_val = (short) (APDU.getBuffer()[tp_SM_CHL]);
    APDU.waitExtension();

    // Length of SMS is CPL - CHL (usually 0x15, see Doc.) - 1 (CHL itself)
    smsLength = (short) ((short) tp_SM_CPL_val - (short) tp_SM_CHL_val - (short) 1);
  }

	public void deselect() {
  }

	public static void install (APDU apdu) {
		new WebSIM();
	}

  /*
  public static void simulate() {
    APDU apdu = new APDU();
    WebSIM hw = new WebSIM(apdu);
    Applet.initSimulator();
    Applet.simulate(hw);
  }

  public static void main(String[] args) {
    WebSIM.simulate();
  }
  */
}

