/*
 * WebSIM - HTTP Server for GSM SIM Card
 *
 * Copyright (c) 1999,2000 SmartCommerce.com, Deutsche Telekom AG
 */

// - Original version by S. Guthery
// - Changed the import list to meet my environment. [H. Vogt 02-12-99]
// - Adapted for use in Simera card [R. Kehr Jan. 2000]

import javacard.framework.*;
import javacardx.framework.*;
import sim.toolkit.*;

public class WebServer extends Applet implements Action, GSMCharacter, Constant {

    WebSIM webSim;

    short counter, itemTextLength,itemCounter, start, len;
    byte[] itemText = new byte[64];
    byte[] bBuffer = new byte[11];

    // Data structures for request handling.
    byte request[];
    short rqp = (short) 0;
    // Pointer to end of request (one byte behind the last byte of the message)
    short rqe;

    // Data structures for response handling.
    byte response[];	// The HTTP response buffer

    byte sessionId;
    byte httpCmd;

    // Here is our magic...
    final public static byte MAGIC_1 = (byte) 0xFA;
    final public static byte MAGIC_2 = (byte) 0xDE;

    private byte ptr;
    private short ef, fs, fe;

    static final byte CHAR_EQUAL_SIGN    = (byte)0x3D;
    static final byte CHAR_QUESTION_MARK = (byte)0x3F;
    static final byte CHAR_AMPERSAND     = (byte)0x26;
    static final byte CHAR_HASH          = (byte)0x23;

    static final byte HTTP_GET  = 0x47;
    static final byte HTTP_HEAD = 0x48;
    static final byte HTTP_POST = 0x50;

    static final short BAD_REQUEST           = 400;
    static final short UNAUTHORIZED          = 401;
    static final short FORBIDDEN             = 403;
    static final short NOT_FOUND             = 404;
    static final short INTERNAL_SERVER_ERROR = 500;
    static final short NOT_IMPLEMENTED       = 501;

    static final short CGIBIN = (short)0xCB00;

    static final byte RESTART_DISP_REQUEST      = (byte) 0x01;
    static final byte RESTART_PROCESS_REQUEST   = (byte) 0x02;
    static final byte RESTART_EXIT              = (byte) 0x03;
    static final byte RESTART_ITEM              = (byte) 0x04;

    static final byte RESTART_SIMAT_SELECT_ITEM = (byte) 0x10;
    static final byte RESTART_SIMAT_GETINPUT    = (byte) 0x11;

    /*
    static byte[] newFileProperties = {
               0x00, 0x00,    // File Length
				       0x00, 0x00,    // File Identifier (FID)
				       0x02,          // Binary File
				       0x01,          // Active
				       0x00,          // Null record length
				       0x00,          // No records
				       (byte)0xFF,    // Default ACL
				       (byte)0x00,    // CHV1 ACL
				       (byte)0x00,    // CHV1 ACL
				       (byte)0xFF,    // AUT0 ACL
				       (byte)0x00,    // AUT1 ACL
				       (byte)0x00,    // AUT2 ACL
				       (byte)0x00,    // AUT3 ACL
				       (byte)0x00 };  // AUT4, ACL

    // These were all static byte arrays before. For some reasons access to these
    // results in an error returning '0000'. Since I made these non-static, access
    // seems to work well.

    // "HTTP/1.0 200 OK"
    static byte[] Status_OK = {(byte)'H', (byte)'T', (byte)'T', (byte)'P', (byte)'/', (byte)'1', (byte)'.', (byte)'0',
			       (byte)' ', (byte)'2', (byte)'0', (byte)'0', (byte)' ', (byte)'O', (byte)'K',
			       (byte)0xD, (byte)0xA};

    // "HTTP/1.0 400 Bad Request"
    static byte[] Status_BR = {(byte)'H', (byte)'T', (byte)'T', (byte)'P', (byte)'/', (byte)'1', (byte)'.', (byte)'0',
			       (byte)' ', (byte)'4', (byte)'0', (byte)'0', (byte)' ', (byte)'B', (byte)'a', (byte)'d',
			       (byte)' ', (byte)'R', (byte)'e', (byte)'q', (byte)'u', (byte)'e', (byte)'s', (byte)'t',
             (byte)0xD, (byte)0xA};

    // "HTTP/1.0 401 Unauthorized"
    static byte[] Status_UA = {(byte)'H', (byte)'T', (byte)'T', (byte)'P', (byte)'/', (byte)'1', (byte)'.', (byte)'0',
			       (byte)' ', (byte)'4', (byte)'0', (byte)'1', (byte)' ', (byte)'U', (byte)'n', (byte)'a',
			       (byte)'u', (byte)'t', (byte)'h', (byte)'o', (byte)'r', (byte)'i', (byte)'z', (byte)'e',
			       (byte)'d', (byte)0xD, (byte)0xA};

    // "HTTP/1.0 403 Forbidden"
    static byte[] Status_FB = {(byte)'H', (byte)'T', (byte)'T', (byte)'P', (byte)'/', (byte)'1', (byte)'.', (byte)'0',
			       (byte)' ', (byte)'4', (byte)'0', (byte)'3', (byte)' ', (byte)'F', (byte)'o', (byte)'r',
			       (byte)'b', (byte)'i', (byte)'d', (byte)'d', (byte)'e', (byte)'n',
			       (byte)0xD, (byte)0xA};

    // "HTTP/1.0 404 Not Found"
    static byte[] Status_NF = {(byte)'H', (byte)'T', (byte)'T', (byte)'P', (byte)'/', (byte)'1', (byte)'.', (byte)'0',
			       (byte)' ', (byte)'4', (byte)'0', (byte)'4', (byte)' ', (byte)'N', (byte)'o', (byte)'t',
			       (byte)'F', (byte)'o', (byte)'u', (byte)'n', (byte)'d',
			       (byte)0xD, (byte)0xA};

    // "HTTP/1.0 500 Internal Server Error"
    static byte[] Status_IE = {(byte)'H', (byte)'T', (byte)'T', (byte)'P', (byte)'/', (byte)'1', (byte)'.', (byte)'0',
			       (byte)' ', (byte)'5', (byte)'0', (byte)'0', (byte)' ', (byte)'I', (byte)'n', (byte)'t',
			       (byte)'e', (byte)'r', (byte)'n', (byte)'a', (byte)'l', (byte)' ', (byte)'S', (byte)'e',
			       (byte)'r', (byte)'v', (byte)'e', (byte)'r', (byte)' ', (byte)'E', (byte)'r', (byte)'r',
			       (byte)'o', (byte)'r', (byte)0xD, (byte)0xA};

    // "HTTP/1.0 501 Not Implemented"
    static byte[] Status_NI = {(byte)'H', (byte)'T', (byte)'T', (byte)'P', (byte)'/', (byte)'1', (byte)'.', (byte)'0',
			       (byte)' ', (byte)'5', (byte)'0', (byte)'1', (byte)' ', (byte)'N', (byte)'o', (byte)'t',
			       (byte)' ', (byte)'I', (byte)'m', (byte)'p', (byte)'l', (byte)'e', (byte)'m', (byte)'e',
			       (byte)'n', (byte)'t', (byte)'e', (byte)'d',
			       (byte)0xD, (byte)0xA};

    // "Server: WebSIM/1.0"
    static byte[] Header_SV = {(byte)'S', (byte)'e', (byte)'r', (byte)'v', (byte)'e', (byte)'r', (byte)':', (byte)' ',
			       (byte)'W', (byte)'e', (byte)'b', (byte)'S', (byte)'I', (byte)'M', (byte)'/', (byte)'1',
			       (byte)'.', (byte)'0',
			       (byte)0xD, (byte)0xA};

    // "Content-Lenght: "
    static byte[] Header_LN = {(byte)'C', (byte)'o', (byte)'n', (byte)'t', (byte)'e', (byte)'n', (byte)'t', (byte)'-',
			       (byte)'L', (byte)'e', (byte)'n', (byte)'g', (byte)'t', (byte)'h', (byte)':', (byte)' ',
			       (byte)' ', (byte)' ', (byte)' ',
			       (byte)0xD, (byte)0xA};

    // "Content-Type: application/octet-stream"
    static byte[] Header_TY = {(byte)'C', (byte)'o', (byte)'n', (byte)'t', (byte)'e', (byte)'n', (byte)'t', (byte)'-',
			       (byte)'T', (byte)'y', (byte)'p', (byte)'e', (byte)':', (byte)' ', (byte)'a', (byte)'p',
			       (byte)'p', (byte)'l', (byte)'i', (byte)'c', (byte)'a', (byte)'t', (byte)'i', (byte)'o',
			       (byte)'n', (byte)'/', (byte)'o', (byte)'c', (byte)'t', (byte)'e', (byte)'t', (byte)'-',
			       (byte)'s', (byte)'t', (byte)'r', (byte)'e', (byte)'a', (byte)'m',
			       (byte)0xD, (byte)0xA};
    */


    // "Received HTTP-Request:"
    //static byte[] ReceivedReq = { R,e,c,e,i,v,e,d, SP, H,T,T,P,(byte)'-',R,e,q,u,e,s,t,(byte)':',SP };
    static byte[] ReceivedReq = { R,e,q,u,e,s,t,(byte)':',SP };

    static byte[] CRLF = { (byte)0xD, (byte)0xA };

    public WebServer() {
      if (request == null) {
        request = new byte[140];  // Not longer than a single SMS message for now
      }
      APDU.waitExtension();
    }


    public void process( WebSIM webSim, byte[] request, short start, short len) {
      this.webSim = webSim;
      // First of all: save the whole request into the temporary buffer
	    Util.arrayCopyNonAtomic( request, start, this.request, (short)0 , (short)len );
      APDU.waitExtension();
      rqe = len;

      //webSim.pushAction( this, RESTART_DISP_REQUEST );
      webSim.pushAction( this, RESTART_PROCESS_REQUEST );
      APDU.waitExtension();
      webSim.playTone.beep();
    }

    public void restart( WebSIM webSim, byte restartAction ) {
      short textLength;
      
      switch( restartAction ) {

        case RESTART_PROCESS_REQUEST:

          // Here we first decide for the command to execute.
          // The actual request is the one displayed previously.
          // Parsing starts at position 3 in the request buffer.
          if (parseHeader() == ST_SUCCESS) {
            rqp = (short) 3;
            parseCommand();
            serve();
          }
          break;

        case RESTART_ITEM:

          if (webSim.selectItem.itemBuf[0] != (byte) 0xFF){
             webSim.currentPage = webSim.smsMessage;
             webSim.pushAction( null, (byte)0 );
             // print selected item (tlv[0] = Item_tag; tlv[1] = length ; tlv[2] = identifier; tlv[3...length-1(?)] = text)
             /*textLength = (short)(webSim.selectItem.itemBuf[1]-(short)1);
             Util.arrayCopyNonAtomic(webSim.selectItem.itemBuf,(short)3,bText,(short)0,textLength);
             webSim.displayMessage.open();
             webSim.displayMessage.writeMessage(bText,(short)0, textLength);
             webSim.displayMessage.writeMessage(webSim.selectItem.itemBuf,(short)3, (short)(webSim.selectItem.itemBuf[(short)1]));
             */
             webSim.smsMessage.setDisplayMessage(webSim.selectItem.itemBuf, (short)3,
                                                (short)(webSim.selectItem.itemBuf[(short)1]-(short)1));
             webSim.smsMessage.setMessage(webSim.selectItem.itemBuf, (short)3,
                                                (short)(webSim.selectItem.itemBuf[(short)1]-(short)1));
             APDU.waitExtension();
             webSim.smsMessage.send();
          }
          break;

        case RESTART_SIMAT_GETINPUT:
          webSim.currentPage = webSim.smsMessage;
          webSim.pushAction( null, (byte)0 );
          /*
          webSim.displayMessage.open();
          webSim.displayMessage.writeByteHex(webSim.input.bLength);
          webSim.displayMessage.writeMessage(webSim.input.bAnswer,(short)0,
                                            (short)webSim.input.bLength);
          webSim.displayMessage.show();
          */
          webSim.smsMessage.setDisplayMessage(webSim.input.bAnswer, (short)0,
                                                (short)webSim.input.bLength);
          webSim.smsMessage.setMessage(webSim.input.bAnswer, (short)0,
                                                (short)webSim.input.bLength);
          APDU.waitExtension();
          webSim.smsMessage.send();
          break;

        case RESTART_EXIT:

          // Do cleanup before exiting the WebServer.
          // Currently there is nothing to do.

          // Do we really need that?
          webSim.pushAction( null, (byte)0 );
          break;

      }
    }

    byte parseHeader() {
      if (rqe <= (short) 3) {
        // Message is too short to be processed.
        return ST_EXIT;
      }
      // Check for magic in request.
      if (request[0] != MAGIC_1 || request[1] != MAGIC_2) {
        // Wrong magic. Ignore this SMS.
        //return ST_EXIT;
      }
      // Fetch session id from request.
      sessionId = request[2];
      return ST_SUCCESS;
    }

    short available() {
      return (short)((short)rqe - (short)rqp);
    }

    void incr( short num ) {
      rqp = (short) ((short)rqp + (short)num);
    }

    void parseCommand() {
      APDU.waitExtension();
      if (available() >= 5) {
        if ( request[rqp]                             == (byte)'G' &&
             request[(short) ((short)rqp + (short)1)] == (byte)'E' &&
             request[(short) ((short)rqp + (short)2)] == (byte)'T' &&
             request[(short) ((short)rqp + (short)3)] == (byte)' ' ) {
           incr( (short)4 );
           httpCmd = HTTP_GET;
        } else if ( request[rqp]                      == (byte)'H' &&
             request[(short) ((short)rqp + (short)1)] == (byte)'E' &&
             request[(short) ((short)rqp + (short)2)] == (byte)'A' &&
             request[(short) ((short)rqp + (short)3)] == (byte)'D' &&
             request[(short) ((short)rqp + (short)4)] == (byte)' ' ) {
           incr( (short)5 );
           httpCmd = HTTP_HEAD;
        } else if ( request[rqp]                      == (byte)'P' &&
             request[(short) ((short)rqp + (short)1)] == (byte)'O' &&
             request[(short) ((short)rqp + (short)2)] == (byte)'S' &&
             request[(short) ((short)rqp + (short)3)] == (byte)'T' &&
             request[(short) ((short)rqp + (short)4)] == (byte)' ' ) {
           incr( (short)5 );
           httpCmd = HTTP_POST;
        } else {
          errorResponse( BAD_REQUEST );
        }
      } else {
        errorResponse( BAD_REQUEST );
      }
    }


    void serve() {
      APDU.waitExtension();

      if (available() >= (short)5) {
        if ( request[rqp]                             == (byte)'/' &&
             request[(short) ((short)rqp + (short)1)] == (byte)'s' &&
             request[(short) ((short)rqp + (short)2)] == (byte)'t' &&
             request[(short) ((short)rqp + (short)3)] == (byte)'/' ) {
          incr( (short)4 );
          processSimATCmd();
        } else if ( request[rqp]                      == (byte)'/' &&
             request[(short) ((short)rqp + (short)1)] == (byte)'f' &&
             request[(short) ((short)rqp + (short)2)] == (byte)'i' &&
             request[(short) ((short)rqp + (short)3)] == (byte)'d' &&
             request[(short) ((short)rqp + (short)4)] == (byte)'/' ) {
          incr( (short)5 );
          //processFileCmd();
        } else {
          errorResponse( NOT_FOUND );
        }
      }
    }

    void processSimATCmd() {

      if (available() >= (short)4) {
        if ( request[rqp]                             == (byte)'s' &&
             request[(short) ((short)rqp + (short)1)] == (byte)'i' &&
             request[(short) ((short)rqp + (short)2)] == (byte)'=' ) {
          // OK, this is the AT Command SELECT ITEM
          incr( (short)3 );
          APDU.waitExtension();
          webSim.pushAction( this, RESTART_SIMAT_SELECT_ITEM );
          APDU.waitExtension();
          //webSim.appSelectItem.process( webSim, request, rqp, (short) ((short)rqe - (short)rqp) );

          start = rqp;
          len = (short) ((short)rqe - (short)rqp);

          webSim.selectItem.newSelection(itemText, (short)0, (short) 0 );
          APDU.waitExtension();
          webSim.currentPage = webSim.selectItem;   //set the correct page

          itemTextLength = (short)1;
          itemCounter = (short) 1;
          // create the different items
          // buf must look like "(<Item1>#<Item2>#...#<ItemN>)" N <= 16 , Length of Itemtext <= 32
          for( counter = (short)(start+(short)1); counter < (start+len); counter = (short)(counter+(short)1)){
            APDU.waitExtension();
            itemText[0] = (byte) itemCounter;

            if ((request[counter] == CHAR_HASH) || (counter == (short)((start+len)-(short)1))) {
              //webSim.selectItem.addItem(itemText, (short)0, (short)itemTextLength);
              if (webSim.selectItem.items[itemCounter] == null) {
                webSim.selectItem.items[itemCounter] = new ProactiveElement(ITEM_TAG, (short)SelectItem.MAXITEMSIZE);
                APDU.waitExtension();
              }
              // webSim.selectItem.items[itemCounter] = new ProactiveElement(ITEM_TAG, (short) 32);

              webSim.selectItem.items[itemCounter].setValue( itemText, (short) 0, (short)32 );
              webSim.selectItem.items[itemCounter].setLength((short)itemTextLength);

              itemTextLength = (short)1;
              itemCounter = (short) (itemCounter + (short) 1);
            } else {
              itemText[itemTextLength] = request[counter];
              itemTextLength = (short) (itemTextLength + (short) 1);
            }
            // save the object to continue after the input
            webSim.pushAction( this,RESTART_ITEM );

            if (webSim.selectItem.buildCommand()== ST_SUCCESS){
              ProactiveCommand.executeCommand();
            }
          }
        }else  if ( request[rqp]                             == (byte)'g' &&
                     ((request[(short) ((short)rqp + (short)1)] == (byte)'i') ||
                     (request[(short) ((short)rqp + (short)1)] == (byte)'n') ||
                     (request[(short) ((short)rqp + (short)1)] == (byte)'p') )&&
                     request[(short) ((short)rqp + (short)2)] == (byte)'=' ) {
          // OK, this is the AT Command GET INPUT
          incr( (short)3 );
          start = rqp;
          len = (short) ((short)rqe - (short)rqp);
          APDU.waitExtension();

          // Save the object to continue after the PCmd Get Input
          webSim.pushAction( this, RESTART_SIMAT_GETINPUT );
          APDU.waitExtension();
          //webSim.input.setText(request, (short)(start+(short)1), (short)(len -(short)1));
          if (request[(short) rqp -(short)2] == (byte)'i'){
            webSim.input.setText(request, (short)(start+(short)1),
                                 (short)(len -(short)1), (byte)ALPHABET );
          } else if  (request[(short) rqp -(short)2] == (byte)'n') {
            webSim.input.setText(request, (short)(start+(short)1),
                                 (short)(len -(short)1), (byte)DIGITS );
          } else {
            webSim.input.setText(request, (short)(start+(short)1),
                                 (short)(len -(short)1),
                                 (byte)((byte)DIGITS | (byte)NOT_REVEALED) );
          }
 
          webSim.currentPage = webSim.input;   //set the correct page
          if (webSim.input.buildCommand()== ST_SUCCESS){
            ProactiveCommand.executeCommand();
          }

        } else if ( request[rqp]                             == (byte)'i' &&
                    request[(short) ((short)rqp + (short)1)] == (byte)'n' &&
                    request[(short) ((short)rqp + (short)2)] == (byte)'f'  &&
                    request[(short) ((short)rqp + (short)3)] == (byte)'o' )  {

          //incr( (short)4 );
          webSim.pushAction( this, RESTART_EXIT );

          webSim.displayMessage.open();
          webSim.displayMessage.writeByte((byte)'S');
          // Enter GSM DF
          if (FileSystem.selectFile((short)0x7F20) == ST.SUCCESS){
            webSim.displayMessage.writeByte((byte)'t');
            APDU.waitExtension();
            // Select EF Loci
            if (FileSystem.selectFile((short)0x6F7E) == ST.SUCCESS){
              webSim.displayMessage.writeByte((byte)'a');
              webSim.displayMessage.writeByte((byte)'t');
              APDU.waitExtension();
              webSim.displayMessage.writeByte((byte)'u');
              webSim.displayMessage.writeByte((byte)'s');
              APDU.waitExtension();
              if (CyberflexOS.readBinaryFile(bBuffer, (short) 0, (short)0,
                                            (short) 0x0B) == ST.SUCCESS ){
                webSim.displayMessage.writeByte((byte)':');
                webSim.displayMessage.writeByte((byte) 0x0A);
                APDU.waitExtension();
                // TMSI
                /*webSim.displayMessage.writeDigit(bBuffer[0]);
                webSim.displayMessage.writeDigit(bBuffer[1]);
                webSim.displayMessage.writeDigit(bBuffer[2]);
                webSim.displayMessage.writeDigit(bBuffer[3]);
                APDU.waitExtension();
                webSim.displayMessage.writeByte((byte) 0x0A);
                */
                // MCC
                webSim.displayMessage.writeByte((byte)'L');
                webSim.displayMessage.writeByte((byte)'A');
                webSim.displayMessage.writeByte((byte)'I');
                APDU.waitExtension();
                webSim.displayMessage.writeByte((byte)':');
                webSim.displayMessage.writeByte(SP);
                webSim.displayMessage.writeDigit(bBuffer[4]);
                webSim.displayMessage.writeDigit(bBuffer[5]);
                APDU.waitExtension();
                webSim.displayMessage.writeByte(SP);
                // MNC
                webSim.displayMessage.writeDigit(bBuffer[6]);
                APDU.waitExtension();
                webSim.displayMessage.writeByte((byte) 0x0A);
                // LAC
                webSim.displayMessage.writeByte((byte)'L');
                webSim.displayMessage.writeByte((byte)'A');
                webSim.displayMessage.writeByte((byte)'C');
                APDU.waitExtension();
                webSim.displayMessage.writeByte((byte)':');
                webSim.displayMessage.writeByte(SP);
                webSim.displayMessage.writeByteHex(bBuffer[7]);
                webSim.displayMessage.writeByteHex(bBuffer[8]);
              }
            }
            webSim.smsMessage.setDisplayMessage(webSim.displayMessage.messageBuf, (short)1,
                                                (short)(webSim.displayMessage.lastChar - (short) 1));
            webSim.smsMessage.setMessage(webSim.displayMessage.messageBuf, (short)1,
                                         (short)(webSim.displayMessage.lastChar - (short) 1));
            APDU.waitExtension();
            webSim.smsMessage.send();
            }
          } else if ( request[rqp]                           == (byte)'d' &&
                    request[(short) ((short)rqp + (short)1)] == (byte)'t' &&
                    request[(short) ((short)rqp + (short)2)] == (byte)'=' )  {

            incr( (short)4 ); // skip '('
            start = rqp;
            len = (short) ((short)rqe - (short)rqp - (short)1); // minus ')'

            Util.arrayCopyNonAtomic( request, (short)start, itemText,
                                     (short)0, (short)len );
            webSim.pushAction( this, RESTART_EXIT );
            webSim.displayMessage.open();
            webSim.displayMessage.writeMessage(itemText, (short)0, len);
            webSim.displayMessage.show();

        } else if ( request[rqp]                         == (byte)'s' &&
                request[(short) ((short)rqp + (short)1)] == (byte)'i' &&
                request[(short) ((short)rqp + (short)2)] == (byte)'g' &&
                request[(short) ((short)rqp + (short)3)] == (byte)'n' &&
                request[(short) ((short)rqp + (short)4)] == (byte)'=')  {

          incr( (short)6 ); // skip '('
          start = rqp;
          len = (short)((short)rqe - (short)rqp - (short)1); // minus ')'
          APDU.waitExtension();

          //Util.arrayCopyNonAtomic(request,(short)start, itemText, (short) 0, (short)(len -(short)1));
          Support.sign( request, start, itemText, (short)0, len );
          webSim.pushAction( this, RESTART_EXIT );

          APDU.waitExtension();
          /*
          webSim.displayMessage.open();
          webSim.displayMessage.writeMessage( request, start, len );
          APDU.waitExtension();
          webSim.displayMessage.writeByte((byte)':');
          APDU.waitExtension();
          webSim.displayMessage.writeMessage( itemText, (short)0, len);
          webSim.displayMessage.show();
          */
          webSim.smsMessage.setDisplayMessage( itemText, (short)0, len );
          webSim.smsMessage.setMessage( itemText, (short)0, len );
          webSim.smsMessage.send();
          }
        }
      }
    /*
    void processFileCmd() {
          webSim.displayMessage.open();
          APDU.waitExtension();
          webSim.displayMessage.writeByte( F );
          webSim.displayMessage.writeByte( i );
          webSim.displayMessage.writeByte( l );
          APDU.waitExtension();
          webSim.displayMessage.writeByte( e );
          APDU.waitExtension();
          webSim.displayMessage.writeByte( (byte) ':' );
          webSim.displayMessage.writeByte( SP );
          APDU.waitExtension();
          webSim.pushAction( this, RESTART_EXIT );
          webSim.displayMessage.show();
     }
     */
    void errorResponse( short statusCode ) {
    }
}
